# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_transfer/transfer.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1apw_transfer/transfer.proto\x12\x0bpw.transfer\"\xfd\x05\n\x05\x43hunk\x12\x13\n\x0btransfer_id\x18\x01 \x01(\r\x12\x1a\n\rpending_bytes\x18\x02 \x01(\rH\x00\x88\x01\x01\x12!\n\x14max_chunk_size_bytes\x18\x03 \x01(\rH\x01\x88\x01\x01\x12#\n\x16min_delay_microseconds\x18\x04 \x01(\rH\x02\x88\x01\x01\x12\x0e\n\x06offset\x18\x05 \x01(\x04\x12\x0c\n\x04\x64\x61ta\x18\x06 \x01(\x0c\x12\x1c\n\x0fremaining_bytes\x18\x07 \x01(\x04H\x03\x88\x01\x01\x12\x13\n\x06status\x18\x08 \x01(\rH\x04\x88\x01\x01\x12\x19\n\x11window_end_offset\x18\t \x01(\r\x12*\n\x04type\x18\n \x01(\x0e\x32\x17.pw.transfer.Chunk.TypeH\x05\x88\x01\x01\x12\x18\n\x0bresource_id\x18\x0b \x01(\rH\x06\x88\x01\x01\x12\x17\n\nsession_id\x18\x0c \x01(\rH\x07\x88\x01\x01\x12\x1d\n\x10protocol_version\x18\r \x01(\rH\x08\x88\x01\x01\x12\x1f\n\x12\x64\x65sired_session_id\x18\x0e \x01(\rH\t\x88\x01\x01\x12\x16\n\x0einitial_offset\x18\x0f \x01(\x04\"\x9e\x01\n\x04Type\x12\x08\n\x04\x44\x41TA\x10\x00\x12\t\n\x05START\x10\x01\x12\x19\n\x15PARAMETERS_RETRANSMIT\x10\x02\x12\x17\n\x13PARAMETERS_CONTINUE\x10\x03\x12\x0e\n\nCOMPLETION\x10\x04\x12\x12\n\x0e\x43OMPLETION_ACK\x10\x05\x12\r\n\tSTART_ACK\x10\x06\x12\x1a\n\x16START_ACK_CONFIRMATION\x10\x07\x42\x10\n\x0e_pending_bytesB\x17\n\x15_max_chunk_size_bytesB\x19\n\x17_min_delay_microsecondsB\x12\n\x10_remaining_bytesB\t\n\x07_statusB\x07\n\x05_typeB\x0e\n\x0c_resource_idB\r\n\x0b_session_idB\x13\n\x11_protocol_versionB\x15\n\x13_desired_session_id\",\n\x15ResourceStatusRequest\x12\x13\n\x0bresource_id\x18\x01 \x01(\r\"\xc6\x01\n\x0eResourceStatus\x12\x13\n\x0bresource_id\x18\x01 \x01(\r\x12\x0e\n\x06status\x18\x02 \x01(\r\x12\x18\n\x10writeable_offset\x18\x03 \x01(\x04\x12\x17\n\x0freadable_offset\x18\x04 \x01(\x04\x12\x1b\n\x0ewrite_checksum\x18\x05 \x01(\x04H\x00\x88\x01\x01\x12\x1a\n\rread_checksum\x18\x06 \x01(\x04H\x01\x88\x01\x01\x42\x11\n\x0f_write_checksumB\x10\n\x0e_read_checksum2\xc9\x01\n\x08Transfer\x12\x32\n\x04Read\x12\x12.pw.transfer.Chunk\x1a\x12.pw.transfer.Chunk(\x01\x30\x01\x12\x33\n\x05Write\x12\x12.pw.transfer.Chunk\x1a\x12.pw.transfer.Chunk(\x01\x30\x01\x12T\n\x11GetResourceStatus\x12\".pw.transfer.ResourceStatusRequest\x1a\x1b.pw.transfer.ResourceStatusB\x1b\n\x17\x64\x65v.pigweed.pw_transferP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_transfer.transfer_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\027dev.pigweed.pw_transferP\001'
  _globals['_CHUNK']._serialized_start=44
  _globals['_CHUNK']._serialized_end=809
  _globals['_CHUNK_TYPE']._serialized_start=466
  _globals['_CHUNK_TYPE']._serialized_end=624
  _globals['_RESOURCESTATUSREQUEST']._serialized_start=811
  _globals['_RESOURCESTATUSREQUEST']._serialized_end=855
  _globals['_RESOURCESTATUS']._serialized_start=858
  _globals['_RESOURCESTATUS']._serialized_end=1056
  _globals['_TRANSFER']._serialized_start=1059
  _globals['_TRANSFER']._serialized_end=1260
# @@protoc_insertion_point(module_scope)
