# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_rpc/benchmark.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16pw_rpc/benchmark.proto\x12\x06pw.rpc\"\x1a\n\x07Payload\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x32u\n\tBenchmark\x12-\n\tUnaryEcho\x12\x0f.pw.rpc.Payload\x1a\x0f.pw.rpc.Payload\x12\x39\n\x11\x42idirectionalEcho\x12\x0f.pw.rpc.Payload\x1a\x0f.pw.rpc.Payload(\x01\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_rpc.benchmark_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_PAYLOAD']._serialized_start=34
  _globals['_PAYLOAD']._serialized_end=60
  _globals['_BENCHMARK']._serialized_start=62
  _globals['_BENCHMARK']._serialized_end=179
# @@protoc_insertion_point(module_scope)
