"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2023 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import pw_chrono_protos.chrono_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class StartRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___StartRequest = StartRequest

@typing_extensions.final
class StartResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___StartResponse = StartResponse

@typing_extensions.final
class StopRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___StopRequest = StopRequest

@typing_extensions.final
class StopResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILE_ID_FIELD_NUMBER: builtins.int
    file_id: builtins.int
    """as a convenience, the file id is returned on stop which can be
    used to start a transfer directly, rather that requiring a user
    list the files to obtain the file id.
    """
    def __init__(
        self,
        *,
        file_id: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_file_id", b"_file_id", "file_id", b"file_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_file_id", b"_file_id", "file_id", b"file_id"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_file_id", b"_file_id"]) -> typing_extensions.Literal["file_id"] | None: ...

global___StopResponse = StopResponse

@typing_extensions.final
class ClockParametersRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___ClockParametersRequest = ClockParametersRequest

@typing_extensions.final
class ClockParametersResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CLOCK_PARAMETERS_FIELD_NUMBER: builtins.int
    @property
    def clock_parameters(self) -> pw_chrono_protos.chrono_pb2.ClockParameters: ...
    def __init__(
        self,
        *,
        clock_parameters: pw_chrono_protos.chrono_pb2.ClockParameters | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["clock_parameters", b"clock_parameters"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["clock_parameters", b"clock_parameters"]) -> None: ...

global___ClockParametersResponse = ClockParametersResponse
