# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_rpc/internal/packet.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cpw_rpc/internal/packet.proto\x12\x0fpw.rpc.internal\"\xa3\x01\n\tRpcPacket\x12)\n\x04type\x18\x01 \x01(\x0e\x32\x1b.pw.rpc.internal.PacketType\x12\x12\n\nchannel_id\x18\x02 \x01(\r\x12\x12\n\nservice_id\x18\x03 \x01(\x07\x12\x11\n\tmethod_id\x18\x04 \x01(\x07\x12\x0f\n\x07payload\x18\x05 \x01(\x0c\x12\x0e\n\x06status\x18\x06 \x01(\r\x12\x0f\n\x07\x63\x61ll_id\x18\x07 \x01(\r*\x9c\x01\n\nPacketType\x12\x0b\n\x07REQUEST\x10\x00\x12\x11\n\rCLIENT_STREAM\x10\x02\x12\x10\n\x0c\x43LIENT_ERROR\x10\x04\x12\x1d\n\x19\x43LIENT_REQUEST_COMPLETION\x10\x08\x12\x0c\n\x08RESPONSE\x10\x01\x12\x10\n\x0cSERVER_ERROR\x10\x05\x12\x11\n\rSERVER_STREAM\x10\x07\"\x04\x08\x03\x10\x03\"\x04\x08\x06\x10\x06\x42\x1d\n\x1b\x64\x65v.pigweed.pw_rpc.internalb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_rpc.internal.packet_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033dev.pigweed.pw_rpc.internal'
  _globals['_PACKETTYPE']._serialized_start=216
  _globals['_PACKETTYPE']._serialized_end=372
  _globals['_RPCPACKET']._serialized_start=50
  _globals['_RPCPACKET']._serialized_end=213
# @@protoc_insertion_point(module_scope)
