# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_trace_protos/trace_rpc.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fpw_trace_protos/trace_rpc.proto\x12\x08pw.trace\"\x07\n\x05\x45mpty\"$\n\x12TraceEnableMessage\x12\x0e\n\x06\x65nable\x18\x01 \x01(\x08\" \n\x10TraceDataMessage\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x32\xd5\x01\n\x0cTraceService\x12\x46\n\x06\x45nable\x12\x1c.pw.trace.TraceEnableMessage\x1a\x1c.pw.trace.TraceEnableMessage\"\x00\x12<\n\tIsEnabled\x12\x0f.pw.trace.Empty\x1a\x1c.pw.trace.TraceEnableMessage\"\x00\x12?\n\x0cGetTraceData\x12\x0f.pw.trace.Empty\x1a\x1a.pw.trace.TraceDataMessage\"\x00\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_trace_protos.trace_rpc_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_EMPTY']._serialized_start=45
  _globals['_EMPTY']._serialized_end=52
  _globals['_TRACEENABLEMESSAGE']._serialized_start=54
  _globals['_TRACEENABLEMESSAGE']._serialized_end=90
  _globals['_TRACEDATAMESSAGE']._serialized_start=92
  _globals['_TRACEDATAMESSAGE']._serialized_end=124
  _globals['_TRACESERVICE']._serialized_start=127
  _globals['_TRACESERVICE']._serialized_end=340
# @@protoc_insertion_point(module_scope)
