# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_i2c/i2c.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10pw_i2c/i2c.proto\x12\x06pw.i2c\"e\n\x0fI2cWriteRequest\x12\x11\n\tbus_index\x18\x01 \x01(\r\x12\x16\n\x0etarget_address\x18\x02 \x01(\r\x12\x18\n\x10register_address\x18\x03 \x01(\x0c\x12\r\n\x05value\x18\x04 \x01(\x0c\"\x12\n\x10I2cWriteResponse\"h\n\x0eI2cReadRequest\x12\x11\n\tbus_index\x18\x01 \x01(\r\x12\x16\n\x0etarget_address\x18\x02 \x01(\r\x12\x18\n\x10register_address\x18\x03 \x01(\x0c\x12\x11\n\tread_size\x18\x04 \x01(\r\" \n\x0fI2cReadResponse\x12\r\n\x05value\x18\x01 \x01(\x0c\x32\x84\x01\n\x03I2c\x12?\n\x08I2cWrite\x12\x17.pw.i2c.I2cWriteRequest\x1a\x18.pw.i2c.I2cWriteResponse\"\x00\x12<\n\x07I2cRead\x12\x16.pw.i2c.I2cReadRequest\x1a\x17.pw.i2c.I2cReadResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_i2c.i2c_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_I2CWRITEREQUEST']._serialized_start=28
  _globals['_I2CWRITEREQUEST']._serialized_end=129
  _globals['_I2CWRITERESPONSE']._serialized_start=131
  _globals['_I2CWRITERESPONSE']._serialized_end=149
  _globals['_I2CREADREQUEST']._serialized_start=151
  _globals['_I2CREADREQUEST']._serialized_end=255
  _globals['_I2CREADRESPONSE']._serialized_start=257
  _globals['_I2CREADRESPONSE']._serialized_end=289
  _globals['_I2C']._serialized_start=292
  _globals['_I2C']._serialized_end=424
# @@protoc_insertion_point(module_scope)
