# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_digital_io/digital_io.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1epw_digital_io/digital_io.proto\x12\rpw.digital_io\"<\n\x16\x44igitalIoEnableRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\x12\x0e\n\x06\x65nable\x18\x02 \x01(\x08\"\x19\n\x17\x44igitalIoEnableResponse\"\\\n\x18\x44igitalIoSetStateRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\x12,\n\x05state\x18\x02 \x01(\x0e\x32\x1d.pw.digital_io.DigitalIoState\"\x1b\n\x19\x44igitalIoSetStateResponse\".\n\x18\x44igitalIoGetStateRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\"I\n\x19\x44igitalIoGetStateResponse\x12,\n\x05state\x18\x01 \x01(\x0e\x32\x1d.pw.digital_io.DigitalIoState**\n\x0e\x44igitalIoState\x12\x0c\n\x08INACTIVE\x10\x00\x12\n\n\x06\x41\x43TIVE\x10\x01\x32\xa8\x02\n\tDigitalIo\x12Y\n\x06\x45nable\x12%.pw.digital_io.DigitalIoEnableRequest\x1a&.pw.digital_io.DigitalIoEnableResponse\"\x00\x12_\n\x08SetState\x12\'.pw.digital_io.DigitalIoSetStateRequest\x1a(.pw.digital_io.DigitalIoSetStateResponse\"\x00\x12_\n\x08GetState\x12\'.pw.digital_io.DigitalIoGetStateRequest\x1a(.pw.digital_io.DigitalIoGetStateResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_digital_io.digital_io_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_DIGITALIOSTATE']._serialized_start=384
  _globals['_DIGITALIOSTATE']._serialized_end=426
  _globals['_DIGITALIOENABLEREQUEST']._serialized_start=49
  _globals['_DIGITALIOENABLEREQUEST']._serialized_end=109
  _globals['_DIGITALIOENABLERESPONSE']._serialized_start=111
  _globals['_DIGITALIOENABLERESPONSE']._serialized_end=136
  _globals['_DIGITALIOSETSTATEREQUEST']._serialized_start=138
  _globals['_DIGITALIOSETSTATEREQUEST']._serialized_end=230
  _globals['_DIGITALIOSETSTATERESPONSE']._serialized_start=232
  _globals['_DIGITALIOSETSTATERESPONSE']._serialized_end=259
  _globals['_DIGITALIOGETSTATEREQUEST']._serialized_start=261
  _globals['_DIGITALIOGETSTATEREQUEST']._serialized_end=307
  _globals['_DIGITALIOGETSTATERESPONSE']._serialized_start=309
  _globals['_DIGITALIOGETSTATERESPONSE']._serialized_end=382
  _globals['_DIGITALIO']._serialized_start=429
  _globals['_DIGITALIO']._serialized_end=725
# @@protoc_insertion_point(module_scope)
