# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_metric_proto/metric_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$pw_metric_proto/metric_service.proto\x12\x0fpw.metric.proto\"`\n\x06Metric\x12\x12\n\ntoken_path\x18\x01 \x03(\x07\x12\x13\n\x0bstring_path\x18\x02 \x03(\t\x12\x12\n\x08\x61s_float\x18\x03 \x01(\x02H\x00\x12\x10\n\x06\x61s_int\x18\x04 \x01(\rH\x00\x42\x07\n\x05value\"9\n\rMetricRequest\x12(\n\x07metrics\x18\x01 \x03(\x0b\x32\x17.pw.metric.proto.Metric\":\n\x0eMetricResponse\x12(\n\x07metrics\x18\x01 \x03(\x0b\x32\x17.pw.metric.proto.Metric2[\n\rMetricService\x12J\n\x03Get\x12\x1e.pw.metric.proto.MetricRequest\x1a\x1f.pw.metric.proto.MetricResponse\"\x00\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_metric_proto.metric_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_METRIC']._serialized_start=57
  _globals['_METRIC']._serialized_end=153
  _globals['_METRICREQUEST']._serialized_start=155
  _globals['_METRICREQUEST']._serialized_end=212
  _globals['_METRICRESPONSE']._serialized_start=214
  _globals['_METRICRESPONSE']._serialized_end=272
  _globals['_METRICSERVICE']._serialized_start=274
  _globals['_METRICSERVICE']._serialized_end=365
# @@protoc_insertion_point(module_scope)
