# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/tuf.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cpw_software_update/tuf.proto\x12\x12pw.software_update\x1a\x1fgoogle/protobuf/timestamp.proto\"i\n\x12SignedRootMetadata\x12 \n\x18serialized_root_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"s\n\x17SignedTimestampMetadata\x12%\n\x1dserialized_timestamp_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"q\n\x16SignedSnapshotMetadata\x12$\n\x1cserialized_snapshot_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"o\n\x15SignedTargetsMetadata\x12#\n\x1bserialized_targets_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"\x91\x01\n\x0e\x43ommonMetadata\x12\x14\n\x0cspec_version\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\r\x12\x30\n\x07\x65xpires\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00\x88\x01\x01\x12\x11\n\x04role\x18\x04 \x01(\tH\x01\x88\x01\x01\x42\n\n\x08_expiresB\x07\n\x05_role\"\xed\x03\n\x0cRootMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12\x1b\n\x13\x63onsistent_snapshot\x18\x02 \x01(\x08\x12,\n\x04keys\x18\x03 \x03(\x0b\x32\x1e.pw.software_update.KeyMapping\x12L\n\x1aroot_signature_requirement\x18\x04 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12Q\n\x1ftimestamp_signature_requirement\x18\x05 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12P\n\x1esnapshot_signature_requirement\x18\x06 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12O\n\x1dtargets_signature_requirement\x18\x07 \x01(\x0b\x32(.pw.software_update.SignatureRequirementJ\x04\x08\x08\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\xa0\x01\n\x11TimestampMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12;\n\x11snapshot_metadata\x18\x02 \x01(\x0b\x32 .pw.software_update.MetadataFileJ\x04\x08\x03\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\x9e\x01\n\x10SnapshotMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12:\n\x10targets_metadata\x18\x02 \x03(\x0b\x32 .pw.software_update.MetadataFileJ\x04\x08\x03\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\x97\x01\n\x0fTargetsMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12\x34\n\x0ctarget_files\x18\x02 \x03(\x0b\x32\x1e.pw.software_update.TargetFileJ\x04\x08\t\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"(\n\tSignature\x12\x0e\n\x06key_id\x18\x01 \x01(\x0c\x12\x0b\n\x03sig\x18\x02 \x01(\x0c\"B\n\nKeyMapping\x12\x0e\n\x06key_id\x18\x01 \x01(\x0c\x12$\n\x03key\x18\x02 \x01(\x0b\x32\x17.pw.software_update.Key\"s\n\x03Key\x12-\n\x08key_type\x18\x01 \x01(\x0e\x32\x1b.pw.software_update.KeyType\x12-\n\x06scheme\x18\x02 \x01(\x0e\x32\x1d.pw.software_update.KeyScheme\x12\x0e\n\x06keyval\x18\x03 \x01(\x0c\":\n\x14SignatureRequirement\x12\x0f\n\x07key_ids\x18\x01 \x03(\x0c\x12\x11\n\tthreshold\x18\x02 \x01(\r\"H\n\x04Hash\x12\x32\n\x08\x66unction\x18\x01 \x01(\x0e\x32 .pw.software_update.HashFunction\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\"l\n\nTargetFile\x12\x11\n\tfile_name\x18\x01 \x01(\t\x12\x0e\n\x06length\x18\x02 \x01(\x04\x12(\n\x06hashes\x18\x03 \x03(\x0b\x32\x18.pw.software_update.HashJ\x04\x08\x04\x10\x10J\x04\x08\x10\x10 J\x05\x08 \x10\x80\x02\"\x8f\x01\n\x0cMetadataFile\x12\x16\n\tfile_name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x0f\n\x07version\x18\x02 \x01(\r\x12\x13\n\x06length\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12(\n\x06hashes\x18\x04 \x03(\x0b\x32\x18.pw.software_update.HashB\x0c\n\n_file_nameB\t\n\x07_length*5\n\x0cHashFunction\x12\x19\n\x15UNKNOWN_HASH_FUNCTION\x10\x00\x12\n\n\x06SHA256\x10\x01*N\n\x07KeyType\x12\x14\n\x10UNKNOWN_KEY_TYPE\x10\x00\x12\x07\n\x03RSA\x10\x01\x12\x0b\n\x07\x45\x44\x32\x35\x35\x31\x39\x10\x02\x12\x17\n\x13\x45\x43\x44SA_SHA2_NISTP256\x10\x03*u\n\tKeyScheme\x12\x16\n\x12UNKNOWN_KEY_SCHEME\x10\x00\x12\x1c\n\x18RSASSA_PSS_SHA256_SCHEME\x10\x01\x12\x12\n\x0e\x45\x44\x32\x35\x35\x31\x39_SCHEME\x10\x02\x12\x1e\n\x1a\x45\x43\x44SA_SHA2_NISTP256_SCHEME\x10\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_software_update.tuf_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_HASHFUNCTION']._serialized_start=2276
  _globals['_HASHFUNCTION']._serialized_end=2329
  _globals['_KEYTYPE']._serialized_start=2331
  _globals['_KEYTYPE']._serialized_end=2409
  _globals['_KEYSCHEME']._serialized_start=2411
  _globals['_KEYSCHEME']._serialized_end=2528
  _globals['_SIGNEDROOTMETADATA']._serialized_start=85
  _globals['_SIGNEDROOTMETADATA']._serialized_end=190
  _globals['_SIGNEDTIMESTAMPMETADATA']._serialized_start=192
  _globals['_SIGNEDTIMESTAMPMETADATA']._serialized_end=307
  _globals['_SIGNEDSNAPSHOTMETADATA']._serialized_start=309
  _globals['_SIGNEDSNAPSHOTMETADATA']._serialized_end=422
  _globals['_SIGNEDTARGETSMETADATA']._serialized_start=424
  _globals['_SIGNEDTARGETSMETADATA']._serialized_end=535
  _globals['_COMMONMETADATA']._serialized_start=538
  _globals['_COMMONMETADATA']._serialized_end=683
  _globals['_ROOTMETADATA']._serialized_start=686
  _globals['_ROOTMETADATA']._serialized_end=1179
  _globals['_TIMESTAMPMETADATA']._serialized_start=1182
  _globals['_TIMESTAMPMETADATA']._serialized_end=1342
  _globals['_SNAPSHOTMETADATA']._serialized_start=1345
  _globals['_SNAPSHOTMETADATA']._serialized_end=1503
  _globals['_TARGETSMETADATA']._serialized_start=1506
  _globals['_TARGETSMETADATA']._serialized_end=1657
  _globals['_SIGNATURE']._serialized_start=1659
  _globals['_SIGNATURE']._serialized_end=1699
  _globals['_KEYMAPPING']._serialized_start=1701
  _globals['_KEYMAPPING']._serialized_end=1767
  _globals['_KEY']._serialized_start=1769
  _globals['_KEY']._serialized_end=1884
  _globals['_SIGNATUREREQUIREMENT']._serialized_start=1886
  _globals['_SIGNATUREREQUIREMENT']._serialized_end=1944
  _globals['_HASH']._serialized_start=1946
  _globals['_HASH']._serialized_end=2018
  _globals['_TARGETFILE']._serialized_start=2020
  _globals['_TARGETFILE']._serialized_end=2128
  _globals['_METADATAFILE']._serialized_start=2131
  _globals['_METADATAFILE']._serialized_end=2274
# @@protoc_insertion_point(module_scope)
