# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_rpc/echo.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11pw_rpc/echo.proto\x12\x06pw.rpc\"\x1a\n\x0b\x45\x63hoMessage\x12\x0b\n\x03msg\x18\x01 \x01(\t2A\n\x0b\x45\x63hoService\x12\x32\n\x04\x45\x63ho\x12\x13.pw.rpc.EchoMessage\x1a\x13.pw.rpc.EchoMessage\"\x00\x42\x1a\n\x18\x64\x65v.pigweed.pw_rpc.protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_rpc.echo_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030dev.pigweed.pw_rpc.proto'
  _globals['_ECHOMESSAGE']._serialized_start=29
  _globals['_ECHOMESSAGE']._serialized_end=55
  _globals['_ECHOSERVICE']._serialized_start=57
  _globals['_ECHOSERVICE']._serialized_end=122
# @@protoc_insertion_point(module_scope)
