# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/update_bundle.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_software_update import tuf_pb2 as pw__software__update_dot_tuf__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&pw_software_update/update_bundle.proto\x12\x12pw.software_update\x1a\x1cpw_software_update/tuf.proto\"\xe5\x04\n\x0cUpdateBundle\x12L\n\x12timestamp_metadata\x18\x01 \x01(\x0b\x32+.pw.software_update.SignedTimestampMetadataH\x00\x88\x01\x01\x12J\n\x11snapshot_metadata\x18\x02 \x01(\x0b\x32*.pw.software_update.SignedSnapshotMetadataH\x01\x88\x01\x01\x12O\n\x10targets_metadata\x18\x03 \x03(\x0b\x32\x35.pw.software_update.UpdateBundle.TargetsMetadataEntry\x12M\n\x0ftarget_payloads\x18\x04 \x03(\x0b\x32\x34.pw.software_update.UpdateBundle.TargetPayloadsEntry\x12\x42\n\rroot_metadata\x18\x05 \x01(\x0b\x32&.pw.software_update.SignedRootMetadataH\x02\x88\x01\x01\x1a\x61\n\x14TargetsMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x38\n\x05value\x18\x02 \x01(\x0b\x32).pw.software_update.SignedTargetsMetadata:\x02\x38\x01\x1a\x35\n\x13TargetPayloadsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x42\x15\n\x13_timestamp_metadataB\x14\n\x12_snapshot_metadataB\x10\n\x0e_root_metadata\"\xe2\x01\n\x08Manifest\x12K\n\x10targets_metadata\x18\x01 \x03(\x0b\x32\x31.pw.software_update.Manifest.TargetsMetadataEntry\x12\x1a\n\ruser_manifest\x18\x02 \x01(\x0cH\x00\x88\x01\x01\x1a[\n\x14TargetsMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.pw.software_update.TargetsMetadata:\x02\x38\x01\x42\x10\n\x0e_user_manifestB\"\n\x1e\x64\x65v.pigweed.pw_software_updateP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_software_update.update_bundle_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\036dev.pigweed.pw_software_updateP\001'
  _globals['_UPDATEBUNDLE_TARGETSMETADATAENTRY']._options = None
  _globals['_UPDATEBUNDLE_TARGETSMETADATAENTRY']._serialized_options = b'8\001'
  _globals['_UPDATEBUNDLE_TARGETPAYLOADSENTRY']._options = None
  _globals['_UPDATEBUNDLE_TARGETPAYLOADSENTRY']._serialized_options = b'8\001'
  _globals['_MANIFEST_TARGETSMETADATAENTRY']._options = None
  _globals['_MANIFEST_TARGETSMETADATAENTRY']._serialized_options = b'8\001'
  _globals['_UPDATEBUNDLE']._serialized_start=93
  _globals['_UPDATEBUNDLE']._serialized_end=706
  _globals['_UPDATEBUNDLE_TARGETSMETADATAENTRY']._serialized_start=491
  _globals['_UPDATEBUNDLE_TARGETSMETADATAENTRY']._serialized_end=588
  _globals['_UPDATEBUNDLE_TARGETPAYLOADSENTRY']._serialized_start=590
  _globals['_UPDATEBUNDLE_TARGETPAYLOADSENTRY']._serialized_end=643
  _globals['_MANIFEST']._serialized_start=709
  _globals['_MANIFEST']._serialized_end=935
  _globals['_MANIFEST_TARGETSMETADATAENTRY']._serialized_start=826
  _globals['_MANIFEST_TARGETSMETADATAENTRY']._serialized_end=917
# @@protoc_insertion_point(module_scope)
