# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_cpu_exception_cortex_m_protos/cpu_state.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0pw_cpu_exception_cortex_m_protos/cpu_state.proto\x12\x19pw.cpu_exception.cortex_m\"\x89\x03\n\x0e\x41rmV7mCpuState\x12\n\n\x02pc\x18\x01 \x01(\r\x12\n\n\x02lr\x18\x02 \x01(\r\x12\x0b\n\x03psr\x18\x03 \x01(\r\x12\x0b\n\x03msp\x18\x04 \x01(\r\x12\x0b\n\x03psp\x18\x05 \x01(\r\x12\x12\n\nexc_return\x18\x06 \x01(\r\x12\x0c\n\x04\x63\x66sr\x18\x07 \x01(\r\x12\x0e\n\x06msplim\x18\x1b \x01(\r\x12\x0e\n\x06psplim\x18\x1c \x01(\r\x12\r\n\x05mmfar\x18\x08 \x01(\r\x12\x0c\n\x04\x62\x66\x61r\x18\t \x01(\r\x12\x0c\n\x04icsr\x18\n \x01(\r\x12\x0c\n\x04hfsr\x18\x19 \x01(\r\x12\r\n\x05shcsr\x18\x1a \x01(\r\x12\x0f\n\x07\x63ontrol\x18\x0b \x01(\r\x12\n\n\x02r0\x18\x0c \x01(\r\x12\n\n\x02r1\x18\r \x01(\r\x12\n\n\x02r2\x18\x0e \x01(\r\x12\n\n\x02r3\x18\x0f \x01(\r\x12\n\n\x02r4\x18\x10 \x01(\r\x12\n\n\x02r5\x18\x11 \x01(\r\x12\n\n\x02r6\x18\x12 \x01(\r\x12\n\n\x02r7\x18\x13 \x01(\r\x12\n\n\x02r8\x18\x14 \x01(\r\x12\n\n\x02r9\x18\x15 \x01(\r\x12\x0b\n\x03r10\x18\x16 \x01(\r\x12\x0b\n\x03r11\x18\x17 \x01(\r\x12\x0b\n\x03r12\x18\x18 \x01(\r\"^\n\x17SnapshotCpuStateOverlay\x12\x43\n\x10\x61rmv7m_cpu_state\x18\x14 \x01(\x0b\x32).pw.cpu_exception.cortex_m.ArmV7mCpuState')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_cpu_exception_cortex_m_protos.cpu_state_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_ARMV7MCPUSTATE']._serialized_start=80
  _globals['_ARMV7MCPUSTATE']._serialized_end=473
  _globals['_SNAPSHOTCPUSTATEOVERLAY']._serialized_start=475
  _globals['_SNAPSHOTCPUSTATEOVERLAY']._serialized_end=569
# @@protoc_insertion_point(module_scope)
