# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_thread_protos/thread.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_tokenizer_proto import options_pb2 as pw__tokenizer__proto_dot_options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dpw_thread_protos/thread.proto\x12\x0fpw.thread.proto\x1a pw_tokenizer_proto/options.proto\"{\n\x0bThreadState\"l\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x15\n\x11INTERRUPT_HANDLER\x10\x01\x12\x0b\n\x07RUNNING\x10\x02\x12\t\n\x05READY\x10\x03\x12\r\n\tSUSPENDED\x10\x04\x12\x0b\n\x07\x42LOCKED\x10\x05\x12\x0c\n\x08INACTIVE\x10\x06\"\xc4\x03\n\x06Thread\x12\x12\n\x04name\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x12\n\x06\x61\x63tive\x18\x02 \x01(\x08\x42\x02\x18\x01\x12\x30\n\x05state\x18\x03 \x01(\x0e\x32!.pw.thread.proto.ThreadState.Enum\x12\x15\n\rraw_backtrace\x18\x04 \x03(\x04\x12\x1c\n\x14symbolized_backtrace\x18\x05 \x03(\t\x12\x11\n\traw_stack\x18\x06 \x01(\x0c\x12 \n\x13stack_start_pointer\x18\x07 \x01(\x04H\x00\x88\x01\x01\x12\x1e\n\x11stack_end_pointer\x18\x08 \x01(\x04H\x01\x88\x01\x01\x12\x1a\n\rstack_pointer\x18\t \x01(\x04H\x02\x88\x01\x01\x12!\n\x14\x63pu_usage_hundredths\x18\n \x01(\rH\x03\x88\x01\x01\x12#\n\x16stack_pointer_est_peak\x18\x0b \x01(\x04H\x04\x88\x01\x01\x42\x16\n\x14_stack_start_pointerB\x14\n\x12_stack_end_pointerB\x10\n\x0e_stack_pointerB\x17\n\x15_cpu_usage_hundredthsB\x19\n\x17_stack_pointer_est_peak\">\n\x12SnapshotThreadInfo\x12(\n\x07threads\x18\x12 \x03(\x0b\x32\x17.pw.thread.proto.ThreadB\x19\n\x0fpw.thread.protoB\x06Threadb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_thread_protos.thread_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\017pw.thread.protoB\006Thread'
  _globals['_THREAD'].fields_by_name['name']._options = None
  _globals['_THREAD'].fields_by_name['name']._serialized_options = b'\200\257&\000'
  _globals['_THREAD'].fields_by_name['active']._options = None
  _globals['_THREAD'].fields_by_name['active']._serialized_options = b'\030\001'
  _globals['_THREADSTATE']._serialized_start=84
  _globals['_THREADSTATE']._serialized_end=207
  _globals['_THREADSTATE_ENUM']._serialized_start=99
  _globals['_THREADSTATE_ENUM']._serialized_end=207
  _globals['_THREAD']._serialized_start=210
  _globals['_THREAD']._serialized_end=662
  _globals['_SNAPSHOTTHREADINFO']._serialized_start=664
  _globals['_SNAPSHOTTHREADINFO']._serialized_end=726
# @@protoc_insertion_point(module_scope)
