# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_snapshot_protos/snapshot.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_chrono_protos import chrono_pb2 as pw__chrono__protos_dot_chrono__pb2
from pw_cpu_exception_cortex_m_protos import cpu_state_pb2 as pw__cpu__exception__cortex__m__protos_dot_cpu__state__pb2
from pw_log.proto import log_pb2 as pw__log_dot_proto_dot_log__pb2
from pw_thread_protos import thread_pb2 as pw__thread__protos_dot_thread__pb2
from pw_snapshot_metadata_proto import snapshot_metadata_pb2 as pw__snapshot__metadata__proto_dot_snapshot__metadata__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!pw_snapshot_protos/snapshot.proto\x12\x0bpw.snapshot\x1a\x1dpw_chrono_protos/chrono.proto\x1a\x30pw_cpu_exception_cortex_m_protos/cpu_state.proto\x1a\x16pw_log/proto/log.proto\x1a\x1dpw_thread_protos/thread.proto\x1a\x32pw_snapshot_metadata_proto/snapshot_metadata.proto\"\xac\x03\n\x08Snapshot\x12\x1e\n\x04logs\x18\x01 \x03(\x0b\x32\x10.pw.log.LogEntry\x12\'\n\x08metadata\x18\x10 \x01(\x0b\x32\x15.pw.snapshot.Metadata\x12-\n\x04tags\x18\x11 \x03(\x0b\x32\x1f.pw.snapshot.Snapshot.TagsEntry\x12(\n\x07threads\x18\x12 \x03(\x0b\x32\x17.pw.thread.proto.Thread\x12\x30\n\x11related_snapshots\x18\x13 \x03(\x0b\x32\x15.pw.snapshot.Snapshot\x12\x43\n\x10\x61rmv7m_cpu_state\x18\x14 \x01(\x0b\x32).pw.cpu_exception.cortex_m.ArmV7mCpuState\x12\x12\n\ntrace_data\x18\x15 \x01(\x0c\x12(\n\ntimestamps\x18\x16 \x03(\x0b\x32\x14.pw.chrono.TimePoint\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01J\x04\x08\x02\x10\x08J\x04\x08\x08\x10\x10J\x05\x08\x17\x10\x88\x08J\t\x08\x88\x08\x10\x80\x80\x80\x80\x02\x42\x1d\n\x11pw.snapshot.protoB\x08Snapshotb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_snapshot_protos.snapshot_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\021pw.snapshot.protoB\010Snapshot'
  _globals['_SNAPSHOT_TAGSENTRY']._options = None
  _globals['_SNAPSHOT_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_SNAPSHOT']._serialized_start=239
  _globals['_SNAPSHOT']._serialized_end=667
  _globals['_SNAPSHOT_TAGSENTRY']._serialized_start=594
  _globals['_SNAPSHOT_TAGSENTRY']._serialized_end=637
# @@protoc_insertion_point(module_scope)
