# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_file/file.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_protobuf_protos import common_pb2 as pw__protobuf__protos_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12pw_file/file.proto\x12\x07pw.file\x1a\x1fpw_protobuf_protos/common.proto\"\x1b\n\x0bListRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\"\x1d\n\rDeleteRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\"\xe5\x01\n\x04Path\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x33\n\x0bpermissions\x18\x02 \x01(\x0e\x32\x19.pw.file.Path.PermissionsH\x00\x88\x01\x01\x12\x17\n\nsize_bytes\x18\x03 \x01(\rH\x01\x88\x01\x01\x12\x14\n\x07\x66ile_id\x18\x04 \x01(\rH\x02\x88\x01\x01\"@\n\x0bPermissions\x12\x08\n\x04NONE\x10\x00\x12\x08\n\x04READ\x10\x01\x12\t\n\x05WRITE\x10\x02\x12\x12\n\x0eREAD_AND_WRITE\x10\x03\x42\x0e\n\x0c_permissionsB\r\n\x0b_size_bytesB\n\n\x08_file_id\",\n\x0cListResponse\x12\x1c\n\x05paths\x18\x01 \x03(\x0b\x32\r.pw.file.Path2}\n\nFileSystem\x12\x37\n\x04List\x12\x14.pw.file.ListRequest\x1a\x15.pw.file.ListResponse\"\x00\x30\x01\x12\x36\n\x06\x44\x65lete\x12\x16.pw.file.DeleteRequest\x1a\x12.pw.protobuf.Empty\"\x00\x42\x06\x42\x04\x46ileb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_file.file_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'B\004File'
  _globals['_LISTREQUEST']._serialized_start=64
  _globals['_LISTREQUEST']._serialized_end=91
  _globals['_DELETEREQUEST']._serialized_start=93
  _globals['_DELETEREQUEST']._serialized_end=122
  _globals['_PATH']._serialized_start=125
  _globals['_PATH']._serialized_end=354
  _globals['_PATH_PERMISSIONS']._serialized_start=247
  _globals['_PATH_PERMISSIONS']._serialized_end=311
  _globals['_LISTRESPONSE']._serialized_start=356
  _globals['_LISTRESPONSE']._serialized_end=400
  _globals['_FILESYSTEM']._serialized_start=402
  _globals['_FILESYSTEM']._serialized_end=527
# @@protoc_insertion_point(module_scope)
