# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_trace_protos/trace.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bpw_trace_protos/trace.proto\x12\x08pw.trace\"\'\n\x11SnapshotTraceInfo\x12\x12\n\ntrace_data\x18\x15 \x01(\x0c\x42\x17\n\x0epw.trace.protoB\x05Traceb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_trace_protos.trace_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\016pw.trace.protoB\005Trace'
  _globals['_SNAPSHOTTRACEINFO']._serialized_start=41
  _globals['_SNAPSHOTTRACEINFO']._serialized_end=80
# @@protoc_insertion_point(module_scope)
