# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_unit_test_proto/unit_test.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"pw_unit_test_proto/unit_test.proto\x12\x0cpw.unit_test\"N\n\x12TestCaseDescriptor\x12\x12\n\nsuite_name\x18\x01 \x01(\t\x12\x11\n\ttest_name\x18\x02 \x01(\t\x12\x11\n\tfile_name\x18\x03 \x01(\t\"m\n\x13TestCaseExpectation\x12\x12\n\nexpression\x18\x01 \x01(\t\x12\x1c\n\x14\x65valuated_expression\x18\x02 \x01(\t\x12\x13\n\x0bline_number\x18\x03 \x01(\r\x12\x0f\n\x07success\x18\x04 \x01(\x08\"\x0e\n\x0cTestRunStart\"O\n\nTestRunEnd\x12\x0e\n\x06passed\x18\x01 \x01(\r\x12\x0e\n\x06\x66\x61iled\x18\x02 \x01(\r\x12\x0f\n\x07skipped\x18\x03 \x01(\r\x12\x10\n\x08\x64isabled\x18\x04 \x01(\r\"\xef\x02\n\x05\x45vent\x12\x34\n\x0etest_run_start\x18\x01 \x01(\x0b\x32\x1a.pw.unit_test.TestRunStartH\x00\x12\x30\n\x0ctest_run_end\x18\x02 \x01(\x0b\x32\x18.pw.unit_test.TestRunEndH\x00\x12;\n\x0ftest_case_start\x18\x03 \x01(\x0b\x32 .pw.unit_test.TestCaseDescriptorH\x00\x12\x35\n\rtest_case_end\x18\x04 \x01(\x0e\x32\x1c.pw.unit_test.TestCaseResultH\x00\x12>\n\x12test_case_disabled\x18\x05 \x01(\x0b\x32 .pw.unit_test.TestCaseDescriptorH\x00\x12\x42\n\x15test_case_expectation\x18\x06 \x01(\x0b\x32!.pw.unit_test.TestCaseExpectationH\x00\x42\x06\n\x04type\"H\n\x0eTestRunRequest\x12\"\n\x1areport_passed_expectations\x18\x01 \x01(\x08\x12\x12\n\ntest_suite\x18\x02 \x03(\t*7\n\x0eTestCaseResult\x12\x0b\n\x07SUCCESS\x10\x00\x12\x0b\n\x07\x46\x41ILURE\x10\x01\x12\x0b\n\x07SKIPPED\x10\x02\x32H\n\x08UnitTest\x12<\n\x03Run\x12\x1c.pw.unit_test.TestRunRequest\x1a\x13.pw.unit_test.Event\"\x00\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_unit_test_proto.unit_test_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_TESTCASERESULT']._serialized_start=784
  _globals['_TESTCASERESULT']._serialized_end=839
  _globals['_TESTCASEDESCRIPTOR']._serialized_start=52
  _globals['_TESTCASEDESCRIPTOR']._serialized_end=130
  _globals['_TESTCASEEXPECTATION']._serialized_start=132
  _globals['_TESTCASEEXPECTATION']._serialized_end=241
  _globals['_TESTRUNSTART']._serialized_start=243
  _globals['_TESTRUNSTART']._serialized_end=257
  _globals['_TESTRUNEND']._serialized_start=259
  _globals['_TESTRUNEND']._serialized_end=338
  _globals['_EVENT']._serialized_start=341
  _globals['_EVENT']._serialized_end=708
  _globals['_TESTRUNREQUEST']._serialized_start=710
  _globals['_TESTRUNREQUEST']._serialized_end=782
  _globals['_UNITTEST']._serialized_start=841
  _globals['_UNITTEST']._serialized_end=913
# @@protoc_insertion_point(module_scope)
