# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_tokenizer_proto/options.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n pw_tokenizer_proto/options.proto\x12\x0cpw.tokenizer\x1a google/protobuf/descriptor.proto*)\n\x0cTokenization\x12\x19\n\x15TOKENIZATION_OPTIONAL\x10\x00:K\n\x06\x66ormat\x12\x1d.google.protobuf.FieldOptions\x18\xf0\xe5\x04 \x01(\x0e\x32\x1a.pw.tokenizer.Tokenizationb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_tokenizer_proto.options_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_TOKENIZATION']._serialized_start=84
  _globals['_TOKENIZATION']._serialized_end=125
# @@protoc_insertion_point(module_scope)
