# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_cpu_exception_risc_v_protos/cpu_state.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.pw_cpu_exception_risc_v_protos/cpu_state.proto\x12\x17pw.cpu_exception.risc_v\"\xab\x03\n\rRiscvCpuState\x12\x0c\n\x04mepc\x18\x01 \x01(\r\x12\x0e\n\x06mcause\x18\x02 \x01(\r\x12\x0f\n\x07mstatus\x18\x03 \x01(\r\x12\r\n\x05mtval\x18\x04 \x01(\r\x12\n\n\x02ra\x18\x05 \x01(\r\x12\n\n\x02sp\x18\x06 \x01(\r\x12\n\n\x02t0\x18\x07 \x01(\r\x12\n\n\x02t1\x18\x08 \x01(\r\x12\n\n\x02t2\x18\t \x01(\r\x12\n\n\x02\x66p\x18\n \x01(\r\x12\n\n\x02s1\x18\x0b \x01(\r\x12\n\n\x02\x61\x30\x18\x0c \x01(\r\x12\n\n\x02\x61\x31\x18\r \x01(\r\x12\n\n\x02\x61\x32\x18\x0e \x01(\r\x12\n\n\x02\x61\x33\x18\x0f \x01(\r\x12\n\n\x02\x61\x34\x18\x10 \x01(\r\x12\n\n\x02\x61\x35\x18\x11 \x01(\r\x12\n\n\x02\x61\x36\x18\x12 \x01(\r\x12\n\n\x02\x61\x37\x18\x1d \x01(\r\x12\n\n\x02s2\x18\x14 \x01(\r\x12\n\n\x02s3\x18\x15 \x01(\r\x12\n\n\x02s4\x18\x16 \x01(\r\x12\n\n\x02s5\x18\x17 \x01(\r\x12\n\n\x02s6\x18\x18 \x01(\r\x12\n\n\x02s7\x18\x19 \x01(\r\x12\n\n\x02s8\x18\x1a \x01(\r\x12\n\n\x02s9\x18\x1b \x01(\r\x12\x0b\n\x03s10\x18\x1c \x01(\r\x12\x0b\n\x03s11\x18\' \x01(\r\x12\n\n\x02t3\x18\x1e \x01(\r\x12\n\n\x02t4\x18\x1f \x01(\r\x12\n\n\x02t5\x18  \x01(\r\x12\n\n\x02t6\x18! \x01(\r\"Z\n\x17SnapshotCpuStateOverlay\x12?\n\x0friscv_cpu_state\x18\x14 \x01(\x0b\x32&.pw.cpu_exception.risc_v.RiscvCpuState')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_cpu_exception_risc_v_protos.cpu_state_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_RISCVCPUSTATE']._serialized_start=76
  _globals['_RISCVCPUSTATE']._serialized_end=503
  _globals['_SNAPSHOTCPUSTATEOVERLAY']._serialized_start=505
  _globals['_SNAPSHOTCPUSTATEOVERLAY']._serialized_end=595
# @@protoc_insertion_point(module_scope)
