# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_thread_protos/thread_snapshot_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_thread_protos import thread_pb2 as pw__thread__protos_dot_thread__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.pw_thread_protos/thread_snapshot_service.proto\x12\x0fpw.thread.proto\x1a\x1dpw_thread_protos/thread.proto\"+\n\rThreadRequest\x12\x11\n\x04name\x18\x01 \x01(\x0cH\x00\x88\x01\x01\x42\x07\n\x05_name2u\n\x15ThreadSnapshotService\x12\\\n\x11GetPeakStackUsage\x12\x1e.pw.thread.proto.ThreadRequest\x1a#.pw.thread.proto.SnapshotThreadInfo\"\x00\x30\x01\x42\x19\n\x0fpw.thread.protoB\x06Threadb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_thread_protos.thread_snapshot_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\017pw.thread.protoB\006Thread'
  _globals['_THREADREQUEST']._serialized_start=98
  _globals['_THREADREQUEST']._serialized_end=141
  _globals['_THREADSNAPSHOTSERVICE']._serialized_start=143
  _globals['_THREADSNAPSHOTSERVICE']._serialized_end=260
# @@protoc_insertion_point(module_scope)
