"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2023 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _DigitalIoState:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _DigitalIoStateEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_DigitalIoState.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    INACTIVE: _DigitalIoState.ValueType  # 0
    ACTIVE: _DigitalIoState.ValueType  # 1

class DigitalIoState(_DigitalIoState, metaclass=_DigitalIoStateEnumTypeWrapper): ...

INACTIVE: DigitalIoState.ValueType  # 0
ACTIVE: DigitalIoState.ValueType  # 1
global___DigitalIoState = DigitalIoState

@typing_extensions.final
class DigitalIoEnableRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LINE_INDEX_FIELD_NUMBER: builtins.int
    ENABLE_FIELD_NUMBER: builtins.int
    line_index: builtins.int
    """Which line to select."""
    enable: builtins.bool
    """Whether to enable or disable."""
    def __init__(
        self,
        *,
        line_index: builtins.int = ...,
        enable: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["enable", b"enable", "line_index", b"line_index"]) -> None: ...

global___DigitalIoEnableRequest = DigitalIoEnableRequest

@typing_extensions.final
class DigitalIoEnableResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___DigitalIoEnableResponse = DigitalIoEnableResponse

@typing_extensions.final
class DigitalIoSetStateRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LINE_INDEX_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    line_index: builtins.int
    """Which line to select."""
    state: global___DigitalIoState.ValueType
    """Which state to set to."""
    def __init__(
        self,
        *,
        line_index: builtins.int = ...,
        state: global___DigitalIoState.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["line_index", b"line_index", "state", b"state"]) -> None: ...

global___DigitalIoSetStateRequest = DigitalIoSetStateRequest

@typing_extensions.final
class DigitalIoSetStateResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___DigitalIoSetStateResponse = DigitalIoSetStateResponse

@typing_extensions.final
class DigitalIoGetStateRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LINE_INDEX_FIELD_NUMBER: builtins.int
    line_index: builtins.int
    """Which line to select."""
    def __init__(
        self,
        *,
        line_index: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["line_index", b"line_index"]) -> None: ...

global___DigitalIoGetStateRequest = DigitalIoGetStateRequest

@typing_extensions.final
class DigitalIoGetStateResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATE_FIELD_NUMBER: builtins.int
    state: global___DigitalIoState.ValueType
    def __init__(
        self,
        *,
        state: global___DigitalIoState.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["state", b"state"]) -> None: ...

global___DigitalIoGetStateResponse = DigitalIoGetStateResponse
