# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_trace_protos/trace_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_chrono_protos import chrono_pb2 as pw__chrono__protos_dot_chrono__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#pw_trace_protos/trace_service.proto\x12\x0epw.trace.proto\x1a\x1dpw_chrono_protos/chrono.proto\"\x0e\n\x0cStartRequest\"\x0f\n\rStartResponse\"\r\n\x0bStopRequest\"0\n\x0cStopResponse\x12\x14\n\x07\x66ile_id\x18\x01 \x01(\rH\x00\x88\x01\x01\x42\n\n\x08_file_id\"\x18\n\x16\x43lockParametersRequest\"O\n\x17\x43lockParametersResponse\x12\x34\n\x10\x63lock_parameters\x18\x01 \x01(\x0b\x32\x1a.pw.chrono.ClockParameters2\x84\x02\n\x0cTraceService\x12\x46\n\x05Start\x12\x1c.pw.trace.proto.StartRequest\x1a\x1d.pw.trace.proto.StartResponse\"\x00\x12\x43\n\x04Stop\x12\x1b.pw.trace.proto.StopRequest\x1a\x1c.pw.trace.proto.StopResponse\"\x00\x12g\n\x12GetClockParameters\x12&.pw.trace.proto.ClockParametersRequest\x1a\'.pw.trace.proto.ClockParametersResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_trace_protos.trace_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_STARTREQUEST']._serialized_start=86
  _globals['_STARTREQUEST']._serialized_end=100
  _globals['_STARTRESPONSE']._serialized_start=102
  _globals['_STARTRESPONSE']._serialized_end=117
  _globals['_STOPREQUEST']._serialized_start=119
  _globals['_STOPREQUEST']._serialized_end=132
  _globals['_STOPRESPONSE']._serialized_start=134
  _globals['_STOPRESPONSE']._serialized_end=182
  _globals['_CLOCKPARAMETERSREQUEST']._serialized_start=184
  _globals['_CLOCKPARAMETERSREQUEST']._serialized_end=208
  _globals['_CLOCKPARAMETERSRESPONSE']._serialized_start=210
  _globals['_CLOCKPARAMETERSRESPONSE']._serialized_end=289
  _globals['_TRACESERVICE']._serialized_start=292
  _globals['_TRACESERVICE']._serialized_end=552
# @@protoc_insertion_point(module_scope)
