# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Tools for managing symlinks."""

import os
from pathlib import Path


def set_symlink(target_path: Path, symlink_path: Path) -> None:
    if symlink_path.exists():
        os.remove(symlink_path)

    os.symlink(target_path, symlink_path, target_path.is_dir())
