# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_snapshot_metadata_proto/snapshot_metadata.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_tokenizer_proto import options_pb2 as pw__tokenizer__proto_dot_options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2pw_snapshot_metadata_proto/snapshot_metadata.proto\x12\x0bpw.snapshot\x1a pw_tokenizer_proto/options.proto\"\x82\x01\n\x0f\x43puArchitecture\"o\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06\x41RMV6M\x10\x01\x12\n\n\x06\x41RMV7M\x10\x02\x12\n\n\x06\x41RMV8M\x10\x03\x12\t\n\x05RV32E\x10\x04\x12\t\n\x05RV32I\x10\x05\x12\t\n\x05RV64E\x10\x06\x12\t\n\x05RV64I\x10\x07\x12\n\n\x06RV128I\x10\x08\"\xe9\x01\n\x08Metadata\x12\x14\n\x06reason\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\r\n\x05\x66\x61tal\x18\x02 \x01(\x08\x12\x1a\n\x0cproject_name\x18\x03 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x18\n\x10software_version\x18\x04 \x01(\t\x12\x1b\n\x13software_build_uuid\x18\x05 \x01(\x0c\x12\x19\n\x0b\x64\x65vice_name\x18\x06 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x15\n\rsnapshot_uuid\x18\x07 \x01(\x0c\x12\x33\n\x08\x63pu_arch\x18\x08 \x01(\x0e\x32!.pw.snapshot.CpuArchitecture.Enum\"\xa1\x01\n\x11SnapshotBasicInfo\x12\'\n\x08metadata\x18\x10 \x01(\x0b\x32\x15.pw.snapshot.Metadata\x12\x36\n\x04tags\x18\x11 \x03(\x0b\x32(.pw.snapshot.SnapshotBasicInfo.TagsEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x1d\n\x11pw.snapshot.protoB\x08Snapshotb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_snapshot_metadata_proto.snapshot_metadata_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\021pw.snapshot.protoB\010Snapshot'
  _globals['_METADATA'].fields_by_name['reason']._options = None
  _globals['_METADATA'].fields_by_name['reason']._serialized_options = b'\200\257&\000'
  _globals['_METADATA'].fields_by_name['project_name']._options = None
  _globals['_METADATA'].fields_by_name['project_name']._serialized_options = b'\200\257&\000'
  _globals['_METADATA'].fields_by_name['device_name']._options = None
  _globals['_METADATA'].fields_by_name['device_name']._serialized_options = b'\200\257&\000'
  _globals['_SNAPSHOTBASICINFO_TAGSENTRY']._options = None
  _globals['_SNAPSHOTBASICINFO_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_CPUARCHITECTURE']._serialized_start=102
  _globals['_CPUARCHITECTURE']._serialized_end=232
  _globals['_CPUARCHITECTURE_ENUM']._serialized_start=121
  _globals['_CPUARCHITECTURE_ENUM']._serialized_end=232
  _globals['_METADATA']._serialized_start=235
  _globals['_METADATA']._serialized_end=468
  _globals['_SNAPSHOTBASICINFO']._serialized_start=471
  _globals['_SNAPSHOTBASICINFO']._serialized_end=632
  _globals['_SNAPSHOTBASICINFO_TAGSENTRY']._serialized_start=589
  _globals['_SNAPSHOTBASICINFO_TAGSENTRY']._serialized_end=632
# @@protoc_insertion_point(module_scope)
