# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import sys
import b
import c
import os
import a

class Tree:
  branches: int
  height_cubits: float

def do_something_really_interesting_with_a_lot_of_arguments(arg_foo,arg_a:str, arg_b:int =0, **kwargs)->None:
       call_some_func()
       print(
         "a str"
       )
