# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_chrono_protos/chrono.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dpw_chrono_protos/chrono.proto\x12\tpw.chrono\"q\n\tEpochType\"d\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x13\n\x0fTIME_SINCE_BOOT\x10\x01\x12\x12\n\x0eUTC_WALL_CLOCK\x10\x02\x12\x12\n\x0eGPS_WALL_CLOCK\x10\x03\x12\x12\n\x0eTAI_WALL_CLOCK\x10\x04\"\xa4\x01\n\x0f\x43lockParameters\x12%\n\x1dtick_period_seconds_numerator\x18\x01 \x01(\x05\x12\'\n\x1ftick_period_seconds_denominator\x18\x02 \x01(\x05\x12\x32\n\nepoch_type\x18\x03 \x01(\x0e\x32\x19.pw.chrono.EpochType.EnumH\x00\x88\x01\x01\x42\r\n\x0b_epoch_type\"T\n\tTimePoint\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\x34\n\x10\x63lock_parameters\x18\x02 \x01(\x0b\x32\x1a.pw.chrono.ClockParameters\">\n\x12SnapshotTimestamps\x12(\n\ntimestamps\x18\x16 \x03(\x0b\x32\x14.pw.chrono.TimePointB\x08\x42\x06\x43hronob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_chrono_protos.chrono_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'B\006Chrono'
  _globals['_EPOCHTYPE']._serialized_start=44
  _globals['_EPOCHTYPE']._serialized_end=157
  _globals['_EPOCHTYPE_ENUM']._serialized_start=57
  _globals['_EPOCHTYPE_ENUM']._serialized_end=157
  _globals['_CLOCKPARAMETERS']._serialized_start=160
  _globals['_CLOCKPARAMETERS']._serialized_end=324
  _globals['_TIMEPOINT']._serialized_start=326
  _globals['_TIMEPOINT']._serialized_end=410
  _globals['_SNAPSHOTTIMESTAMPS']._serialized_start=412
  _globals['_SNAPSHOTTIMESTAMPS']._serialized_end=474
# @@protoc_insertion_point(module_scope)
