# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_protobuf_codegen_protos/codegen_options.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0pw_protobuf_codegen_protos/codegen_options.proto\x12\x0bpw.protobuf\"t\n\x0e\x43odegenOptions\x12\x11\n\tmax_count\x18\x01 \x01(\x05\x12\x13\n\x0b\x66ixed_count\x18\x02 \x01(\x08\x12\x10\n\x08max_size\x18\x03 \x01(\x05\x12\x12\n\nfixed_size\x18\x04 \x01(\x08\x12\x14\n\x0cuse_callback\x18\x05 \x01(\x08\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_protobuf_codegen_protos.codegen_options_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_CODEGENOPTIONS']._serialized_start=65
  _globals['_CODEGENOPTIONS']._serialized_end=181
# @@protoc_insertion_point(module_scope)
