# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_log/proto/log.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_protobuf_protos import common_pb2 as pw__protobuf__protos_dot_common__pb2
from pw_tokenizer_proto import options_pb2 as pw__tokenizer__proto_dot_options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16pw_log/proto/log.proto\x12\x06pw.log\x1a\x1fpw_protobuf_protos/common.proto\x1a pw_tokenizer_proto/options.proto\"\xe4\x01\n\x08LogEntry\x12\x1a\n\x07message\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00H\x01\x88\x01\x01\x12\x12\n\nline_level\x18\x02 \x01(\r\x12\r\n\x05\x66lags\x18\x03 \x01(\r\x12\x13\n\ttimestamp\x18\x04 \x01(\x03H\x00\x12\x1f\n\x15time_since_last_entry\x18\x05 \x01(\x03H\x00\x12\x0f\n\x07\x64ropped\x18\x06 \x01(\r\x12\x14\n\x06module\x18\x07 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x12\n\x04\x66ile\x18\x08 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x14\n\x06thread\x18\t \x01(\x0c\x42\x04\x80\xaf&\x00\x42\x06\n\x04timeB\n\n\x08_message\"\x0c\n\nLogRequest\"P\n\nLogEntries\x12!\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x10.pw.log.LogEntry\x12\x1f\n\x17\x66irst_entry_sequence_id\x18\x02 \x01(\r\"\xf2\x02\n\nFilterRule\x12=\n\x1blevel_greater_than_or_equal\x18\x01 \x01(\x0e\x32\x18.pw.log.FilterRule.Level\x12\x1b\n\rmodule_equals\x18\x02 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x15\n\rany_flags_set\x18\x03 \x01(\r\x12)\n\x06\x61\x63tion\x18\x04 \x01(\x0e\x32\x19.pw.log.FilterRule.Action\x12\x1b\n\rthread_equals\x18\x05 \x01(\x0c\x42\x04\x80\xaf&\x00\"}\n\x05Level\x12\r\n\tANY_LEVEL\x10\x00\x12\x0f\n\x0b\x44\x45\x42UG_LEVEL\x10\x01\x12\x0e\n\nINFO_LEVEL\x10\x02\x12\x0e\n\nWARN_LEVEL\x10\x03\x12\x0f\n\x0b\x45RROR_LEVEL\x10\x04\x12\x12\n\x0e\x43RITICAL_LEVEL\x10\x05\x12\x0f\n\x0b\x46\x41TAL_LEVEL\x10\x07\"*\n\x06\x41\x63tion\x12\x0c\n\x08INACTIVE\x10\x00\x12\x08\n\x04KEEP\x10\x01\x12\x08\n\x04\x44ROP\x10\x02\"*\n\x06\x46ilter\x12 \n\x04rule\x18\x01 \x03(\x0b\x32\x12.pw.log.FilterRule\"K\n\x10SetFilterRequest\x12\x17\n\tfilter_id\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x1e\n\x06\x66ilter\x18\x02 \x01(\x0b\x32\x0e.pw.log.Filter\"+\n\x10GetFilterRequest\x12\x17\n\tfilter_id\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\"\x15\n\x13\x46ilterIdListRequest\"/\n\x14\x46ilterIdListResponse\x12\x17\n\tfilter_id\x18\x01 \x03(\x0c\x42\x04\x80\xaf&\x00\x32:\n\x04Logs\x12\x32\n\x06Listen\x12\x12.pw.log.LogRequest\x1a\x12.pw.log.LogEntries0\x01\x32\xc7\x01\n\x07\x46ilters\x12\x39\n\tSetFilter\x12\x18.pw.log.SetFilterRequest\x1a\x12.pw.protobuf.Empty\x12\x35\n\tGetFilter\x12\x18.pw.log.GetFilterRequest\x1a\x0e.pw.log.Filter\x12J\n\rListFilterIds\x12\x1b.pw.log.FilterIdListRequest\x1a\x1c.pw.log.FilterIdListResponseB\x05\x42\x03Logb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_log.proto.log_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'B\003Log'
  _globals['_LOGENTRY'].fields_by_name['message']._options = None
  _globals['_LOGENTRY'].fields_by_name['message']._serialized_options = b'\200\257&\000'
  _globals['_LOGENTRY'].fields_by_name['module']._options = None
  _globals['_LOGENTRY'].fields_by_name['module']._serialized_options = b'\200\257&\000'
  _globals['_LOGENTRY'].fields_by_name['file']._options = None
  _globals['_LOGENTRY'].fields_by_name['file']._serialized_options = b'\200\257&\000'
  _globals['_LOGENTRY'].fields_by_name['thread']._options = None
  _globals['_LOGENTRY'].fields_by_name['thread']._serialized_options = b'\200\257&\000'
  _globals['_FILTERRULE'].fields_by_name['module_equals']._options = None
  _globals['_FILTERRULE'].fields_by_name['module_equals']._serialized_options = b'\200\257&\000'
  _globals['_FILTERRULE'].fields_by_name['thread_equals']._options = None
  _globals['_FILTERRULE'].fields_by_name['thread_equals']._serialized_options = b'\200\257&\000'
  _globals['_SETFILTERREQUEST'].fields_by_name['filter_id']._options = None
  _globals['_SETFILTERREQUEST'].fields_by_name['filter_id']._serialized_options = b'\200\257&\000'
  _globals['_GETFILTERREQUEST'].fields_by_name['filter_id']._options = None
  _globals['_GETFILTERREQUEST'].fields_by_name['filter_id']._serialized_options = b'\200\257&\000'
  _globals['_FILTERIDLISTRESPONSE'].fields_by_name['filter_id']._options = None
  _globals['_FILTERIDLISTRESPONSE'].fields_by_name['filter_id']._serialized_options = b'\200\257&\000'
  _globals['_LOGENTRY']._serialized_start=102
  _globals['_LOGENTRY']._serialized_end=330
  _globals['_LOGREQUEST']._serialized_start=332
  _globals['_LOGREQUEST']._serialized_end=344
  _globals['_LOGENTRIES']._serialized_start=346
  _globals['_LOGENTRIES']._serialized_end=426
  _globals['_FILTERRULE']._serialized_start=429
  _globals['_FILTERRULE']._serialized_end=799
  _globals['_FILTERRULE_LEVEL']._serialized_start=630
  _globals['_FILTERRULE_LEVEL']._serialized_end=755
  _globals['_FILTERRULE_ACTION']._serialized_start=757
  _globals['_FILTERRULE_ACTION']._serialized_end=799
  _globals['_FILTER']._serialized_start=801
  _globals['_FILTER']._serialized_end=843
  _globals['_SETFILTERREQUEST']._serialized_start=845
  _globals['_SETFILTERREQUEST']._serialized_end=920
  _globals['_GETFILTERREQUEST']._serialized_start=922
  _globals['_GETFILTERREQUEST']._serialized_end=965
  _globals['_FILTERIDLISTREQUEST']._serialized_start=967
  _globals['_FILTERIDLISTREQUEST']._serialized_end=988
  _globals['_FILTERIDLISTRESPONSE']._serialized_start=990
  _globals['_FILTERIDLISTRESPONSE']._serialized_end=1037
  _globals['_LOGS']._serialized_start=1039
  _globals['_LOGS']._serialized_end=1097
  _globals['_FILTERS']._serialized_start=1100
  _globals['_FILTERS']._serialized_end=1299
# @@protoc_insertion_point(module_scope)
