# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_protobuf_protos/field_options.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&pw_protobuf_protos/field_options.proto\x12\x0bpw.protobuf\x1a google/protobuf/descriptor.proto\"X\n\x0c\x46ieldOptions\x12\x16\n\tmax_count\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12\x15\n\x08max_size\x18\x02 \x01(\x05H\x01\x88\x01\x01\x42\x0c\n\n_max_countB\x0b\n\t_max_size:J\n\x04pwpb\x12\x1d.google.protobuf.FieldOptions\x18\x83\t \x01(\x0b\x32\x19.pw.protobuf.FieldOptions\x88\x01\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_protobuf_protos.field_options_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_FIELDOPTIONS']._serialized_start=89
  _globals['_FIELDOPTIONS']._serialized_end=177
# @@protoc_insertion_point(module_scope)
