# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_system_protos/device_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%pw_system_protos/device_service.proto\x12\x0fpw.system.proto\"\x0f\n\rRebootRequest\"\x10\n\x0eRebootResponse\"\x0e\n\x0c\x43rashRequest\"\x0f\n\rCrashResponse2\xa6\x01\n\rDeviceService\x12K\n\x06Reboot\x12\x1e.pw.system.proto.RebootRequest\x1a\x1f.pw.system.proto.RebootResponse\"\x00\x12H\n\x05\x43rash\x12\x1d.pw.system.proto.CrashRequest\x1a\x1e.pw.system.proto.CrashResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_system_protos.device_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_REBOOTREQUEST']._serialized_start=58
  _globals['_REBOOTREQUEST']._serialized_end=73
  _globals['_REBOOTRESPONSE']._serialized_start=75
  _globals['_REBOOTRESPONSE']._serialized_end=91
  _globals['_CRASHREQUEST']._serialized_start=93
  _globals['_CRASHREQUEST']._serialized_end=107
  _globals['_CRASHRESPONSE']._serialized_start=109
  _globals['_CRASHRESPONSE']._serialized_end=124
  _globals['_DEVICESERVICE']._serialized_start=127
  _globals['_DEVICESERVICE']._serialized_end=293
# @@protoc_insertion_point(module_scope)
