# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/bundled_update.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from pw_protobuf_protos import common_pb2 as pw__protobuf__protos_dot_common__pb2
from pw_tokenizer_proto import options_pb2 as pw__tokenizer__proto_dot_options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'pw_software_update/bundled_update.proto\x12\x12pw.software_update\x1a\x19google/protobuf/any.proto\x1a\x1fpw_protobuf_protos/common.proto\x1a pw_tokenizer_proto/options.proto\"\xa2\x01\n\x12\x42undledUpdateState\"\x8b\x01\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0c\n\x08INACTIVE\x10\x01\x12\x10\n\x0cTRANSFERRING\x10\x02\x12\x0f\n\x0bTRANSFERRED\x10\x03\x12\r\n\tVERIFYING\x10\x04\x12\x0c\n\x08VERIFIED\x10\x05\x12\x0c\n\x08\x41PPLYING\x10\x06\x12\x0c\n\x08\x41\x42ORTING\x10\x07\x12\x0c\n\x08\x46INISHED\x10\x08\"\x91\x01\n\x13\x42undledUpdateResult\"z\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07SUCCESS\x10\x01\x12\x11\n\rUNKNOWN_ERROR\x10\x02\x12\x0b\n\x07\x41\x42ORTED\x10\x03\x12\x13\n\x0fTRANSFER_FAILED\x10\x04\x12\x11\n\rVERIFY_FAILED\x10\x05\x12\x10\n\x0c\x41PPLY_FAILED\x10\x06\"\xc7\x03\n\x13\x42undledUpdateStatus\x12:\n\x05state\x18\x01 \x01(\x0e\x32+.pw.software_update.BundledUpdateState.Enum\x12\x41\n\x06result\x18\x02 \x01(\x0e\x32,.pw.software_update.BundledUpdateResult.EnumH\x00\x88\x01\x01\x12\x34\n\'current_state_progress_hundreth_percent\x18\x03 \x01(\rH\x01\x88\x01\x01\x12\x18\n\x0btransfer_id\x18\x04 \x01(\rH\x02\x88\x01\x01\x12\x1c\n\x0f\x62undle_filename\x18\x05 \x01(\tH\x03\x88\x01\x01\x12\x17\n\x04note\x18\x06 \x01(\x0c\x42\x04\x80\xaf&\x00H\x04\x88\x01\x01\x12\x32\n\x0f\x65xtended_status\x18\x07 \x01(\x0b\x32\x14.google.protobuf.AnyH\x05\x88\x01\x01\x42\t\n\x07_resultB*\n(_current_state_progress_hundreth_percentB\x0e\n\x0c_transfer_idB\x12\n\x10_bundle_filenameB\x07\n\x05_noteB\x12\n\x10_extended_status\"@\n\x0cStartRequest\x12\x1c\n\x0f\x62undle_filename\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x12\n\x10_bundle_filename2\x95\x04\n\rBundledUpdate\x12H\n\tGetStatus\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12R\n\x05Start\x12 .pw.software_update.StartRequest\x1a\'.pw.software_update.BundledUpdateStatus\x12M\n\x0eSetTransferred\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x45\n\x06Verify\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05\x41pply\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05\x41\x62ort\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05Reset\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatusB\"\n\x1e\x64\x65v.pigweed.pw_software_updateP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_software_update.bundled_update_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\036dev.pigweed.pw_software_updateP\001'
  _globals['_BUNDLEDUPDATESTATUS'].fields_by_name['note']._options = None
  _globals['_BUNDLEDUPDATESTATUS'].fields_by_name['note']._serialized_options = b'\200\257&\000'
  _globals['_BUNDLEDUPDATESTATE']._serialized_start=158
  _globals['_BUNDLEDUPDATESTATE']._serialized_end=320
  _globals['_BUNDLEDUPDATESTATE_ENUM']._serialized_start=181
  _globals['_BUNDLEDUPDATESTATE_ENUM']._serialized_end=320
  _globals['_BUNDLEDUPDATERESULT']._serialized_start=323
  _globals['_BUNDLEDUPDATERESULT']._serialized_end=468
  _globals['_BUNDLEDUPDATERESULT_ENUM']._serialized_start=346
  _globals['_BUNDLEDUPDATERESULT_ENUM']._serialized_end=468
  _globals['_BUNDLEDUPDATESTATUS']._serialized_start=471
  _globals['_BUNDLEDUPDATESTATUS']._serialized_end=926
  _globals['_STARTREQUEST']._serialized_start=928
  _globals['_STARTREQUEST']._serialized_end=992
  _globals['_BUNDLEDUPDATE']._serialized_start=995
  _globals['_BUNDLEDUPDATE']._serialized_end=1528
# @@protoc_insertion_point(module_scope)
